//----------------------------------------------------------------------------
// 
//	file: simenv.cc
// 
//	function: simulation environment and scheduling, etc. 
//	author:   Yoonsuck Choe <yschoe@cs.utexas.edu>
//	creation: Sat Oct 10 04:51:55 CDT 1998
// 
//	$Revision: 1.5 $ $Date: 2001/11/30 20:24:15 $
//	$Log: simenv.cc,v $
//	Revision 1.5  2001/11/30 20:24:15  yschoe
//	pglissom v1.0 release
//
//	Revision 1.4  2000/01/30 14:56:21  yschoe
//	*** empty log message ***
//
//	Revision 1.3  1999/02/09 07:42:25  yschoe
//	1. alias setting allowed
//	2. more stuff printed when status is asked
//
//	Revision 1.2  1998/10/26 03:54:11  yschoe
//	1. added hook counts for different kinds of hooks
//	2. added hook tables for the same
//	3. check for command redundancy when adding command
//	4. implemented all supported actions
//	5. added parse_file() for reading and executing commands from a file
//	   directly
//	6. added list_hooks() for a list of hooks
//	7. ** added rerun_hooks() : to be used as a preparation stage for
//	   loading a network image (in netlayer class)
//
//	Revision 1.1  1998/10/16 10:08:04  yschoe
//	Initial revision
//
//
//----------------------------------------------------------------------------

#include "simenv.h"

//----------------------------------------------------------------------------
//	simenv() 
//	  - Constructor
//----------------------------------------------------------------------------
simenv::simenv(char *comment, int max) : blackboard(comment, max) {

  // initialize counters
  obj_count = 0;
  cmd_count = 0;
  hook_init_count = 0;
  hook_after_count = 0;
  hook_before_count = 0;

  // setup command and object list
  obj_table 	= new SIM_OBJ [max];
  cmd_table 	= new SIM_CMD [max];
  hook_init_table = new SIM_HOOK [max];
  hook_after_table = new SIM_HOOK [max];
  hook_before_table = new SIM_HOOK [max];

  // malloc sim_argv vector
  for (int i=0; i<SIM_MAX_CMD_ARGS; ++i) {
	sim_argv[i] = new char [SIM_ARG_LEN];
  }

}
  
//----------------------------------------------------------------------------
//	~simenv() 
//	  - Destructor
//----------------------------------------------------------------------------
simenv::~simenv() {

  // dealloc tables created in here
  delete [] obj_table;
  delete [] cmd_table;

  // dealloc sim_argv
  for (int i=0; i<SIM_MAX_CMD_ARGS; ++i) {
	delete [] sim_argv[i];
  }

}

//----------------------------------------------------------------------------
//	add_obj() 
//	  - add an object (instance of other classes) to obj_table
//----------------------------------------------------------------------------
void simenv::add_obj(char *name, netlayer *obj, char *comment) {

  // No redundancy check is done!

  SIM_OBJ *objp = &(obj_table[obj_count]);
  init_obj(name,objp);
  objp->objp = obj;
  obj_count++;

}

//----------------------------------------------------------------------------
//	add_cmd() 
//	  - add a command (member function of a specific object)
//	  - the object in question must have been added already
//	  - the function that can be added is quite limited in that it
//	    must only take a list of strings as argument, and return 
//	    an integer
//----------------------------------------------------------------------------
void simenv::add_cmd(char *cmd_name, char *obj_name,    
                SIM_MEMBER_FUNCP function, int num_args, char *help_string) {

  // check if command already exists
  if (find_cmd(cmd_name)>0) {
	warning("simenv:add_cmd","command \"%s\" already exists",cmd_name);
	return;
  }	

  // find and check if an object by the name 'obj_name' exists.
  int obj_idx = find_obj(obj_name);
  if (obj_idx < 0) {
        warning("simenv:add_cmd","oject \"%s\" does not exist for command %s",
		obj_name, cmd_name);
	return;
  }

  SIM_CMD *cmd = &(cmd_table[cmd_count]);
  init_cmd(cmd_name,cmd,num_args,help_string);
  cmd->objp  = obj_table[obj_idx].objp;
  cmd->member_funcp = function;
  cmd->funcp = (SIM_FUNCP)NULL;
  cmd_count++;

}
//----------------------------------------------------------------------------
// add regular C function that returns an integer
void simenv::add_cmd(char *cmd_name,
                SIM_FUNCP function, int num_args, char *help_string) {

  // check if command already exists
  if (find_cmd(cmd_name)>0) {
	warning("simenv:add_cmd","command \"%s\" already exists",cmd_name);
  }	

  cout << "simenv:add_cmd() : adding command " << cmd_name << endl;
	
  SIM_CMD *cmd = &(cmd_table[cmd_count]);
  init_cmd(cmd_name,cmd,num_args,help_string);
  cmd->objp  = (SIM_OBJP)NULL;
  cmd->member_funcp = (SIM_MEMBER_FUNCP)NULL;
  cmd->funcp = function;
  cmd_count++;

}

//----------------------------------------------------------------------------
//	add_alias() 
//	  - make 'alias' the alias of 'original'
//----------------------------------------------------------------------------
void simenv::add_alias(char *alias, char *original) {

  // This is actually just adding another obj with the same
  // content as the original

  // No redundancy check is done!
  int original_idx = find_obj(original);

  // Get the obj_table index of the original
  SIM_OBJ *origp = &(obj_table[original_idx]);

  // Now add a new obj under the aliased name
  SIM_OBJ *aliasp = &(obj_table[obj_count]); 
  init_obj(alias,aliasp);
  aliasp->objp = origp->objp;
  obj_count++;

}

//----------------------------------------------------------------------------
//	parse() 
//	  - parse and either run or save it in hook list
//----------------------------------------------------------------------------
void simenv::parse(char *command) {

  // pad eoln with ' ' for proper parsing of single token commands
  if (command[strlen(command)-1]=='\n') {
	command[strlen(command)-1] = '\0';	
  }

  if (strlen(command)==0) return;

  // * if sim_parser->match("pattern") did not generate a match,
  //   it restores the original command. Otherwise, the residual
  //   part of the string (less the original pattern) remains in 
  //   the string command.

  // 0. set global parameter : parameters registered to *this.
  //	Use the one already implemented in blackboard class
  if (set_parameter(command)) {
	// action is already taken above
	return;
  } // done

  // start up parser
  sim_parser = new parser(command); 

  // 1. comment starts with '#" 
  if (sim_parser->match("#")) {  // done
	// do nothing (comment)
  } else 
  // 2. set parameter specific to a registered object
  if (sim_parser->match("obj")) { // done
  	set_objparam();
  } else 
  // 3. set hook
  if (sim_parser->match("hook")) { // done
  	set_hook();
  } else 
  // 4. help 
  if (sim_parser->match("help")) {
	help();
  } else 
  // 5. run command
  if (sim_parser->match("run") ) { // postpone 10/17
	run();
  } else 
  // 6. kill command : kill program
  if (sim_parser->match("kill") ) { 
	cerr << "simenv -- kill command encounterd. Killing application" 
		<< endl;
	cerr << " * if program hangs, type CTRL-Z and kill the job manually"
		<< endl;
	exit(0);
  } else {
	cerr << "simenv::parse() : unknown command in \""
	     << command << "\"" << endl;
  } 
  // sim_parser should be deleted after processing each command string.

  delete sim_parser;
  return;

}


//----------------------------------------------------------------------------
//	parse_file() 
//	  - read lines from file and call parse()
//----------------------------------------------------------------------------
int   simenv::parse_file(char *filename) {

  char buff[SIM_STR_LEN];
  int  count = 0;

  FILE *conffp = fopen(filename,"r");

  if (conffp==NULL) {
	cerr << "simenv::parse_file: cannot open file " << filename << endl;
	exit(0);
  }

  while (!feof(conffp)) {

        fgets(buff,SIM_STR_LEN,conffp);
        // if (feof(conffp)) { break; }
        //parse((char*)&(buff[0]));
	parse(buff);
        count ++;

  }

  return count;

}

//----------------------------------------------------------------------------
//	set_objparam() 
//	  - set parameter in a certain registered object
//----------------------------------------------------------------------------
void simenv::set_objparam() {

  // Assume that the object in question has a blackboard 
  // member, and all the 

  // Allowed format:
  //            obj <obj> set <parameter> = <value>
  //            obj <obj> set <parameter>[<column>] = <value>
  //            obj <obj> set <parameter>[<col>][<row>] = <value>

  // 1. get object name
  char obj_name[SIM_STR_LEN];
  char cmd_str[SIM_STR_LEN];
  char cmd_bak[SIM_STR_LEN];
  int  col, row;
  int  obj_idx;

  if (!sim_parser->match_param(obj_name,&col,&row)) {
	warning("simenv::set_objparam","syntax error : %s",
		obj_name);
 	return;
  }

  // cout << "obj_name = " << obj_name << endl;

  if ((obj_idx=find_obj(obj_name))<0) {
	warning("simenv::set_objparam","unknown object \"%s\"",obj_name);
	return;
  }

  // get the rest of the command string and pass it over.
  obj_table[obj_idx].objp->set_parameter(sim_parser);

}

//----------------------------------------------------------------------------
//	run() 
//	  - run registered command
//----------------------------------------------------------------------------
void simenv::run() {

  char arg_val[SIM_STR_LEN];
  
  // 1. parse the command name 
  if (!sim_parser->next(sim_argv[0])) {
	warning("simenv::run","syntax error");
	return;
  }

  // 2. find command in command table
  int cmd_idx = find_cmd(sim_argv[0]);
  if (cmd_idx<0) {
	warning("simenv::run","command \"%s\" not found in command table\n",
		sim_argv[0]);
  	return;
  }

  // 3. seperate arguments from the arg string and put it in argv[]
  for (int i = 1; i <= cmd_table[cmd_idx].num_args; ++i) {
	if (!sim_parser->next(sim_argv[i])) {
	   warning("simenv::run","argument missing for \"%s\"",sim_argv[0]);
           return;
	} 
  }

  // 4. call function
  if (cmd_table[cmd_idx].objp==NULL) {
	(*cmd_table[cmd_idx].funcp)(cmd_table[cmd_idx].num_args,sim_argv);
  } else {
	CALL_MEMBER_FUNC(
		(cmd_table[cmd_idx].objp),
		(cmd_table[cmd_idx].member_funcp)
        )(cmd_table[cmd_idx].num_args,sim_argv);
  }

  /*
  cout << "command: " << sim_argv[0] << " idx=" << cmd_idx << endl;
  for (int i=1; i<=cmd_table[cmd_idx].num_args; ++i) {
	cout << "arg " << i << " = " << sim_argv[i] << endl;
  }
  */

  return;

}

//----------------------------------------------------------------------------
//	set_hook() 
//	  - set command hook to run at a certain time before other routines
//----------------------------------------------------------------------------
void simenv::set_hook() {

  //
  // Depending on what run_hook function should execute it
  //
  // 1. run_hook_init() 
  if (sim_parser->match("init")) {
	hook_init_table[hook_init_count].time = 0;
	sim_parser->rest(hook_init_table[hook_init_count].command);
	hook_init_table[hook_init_count].exec_flag = SIM_HOOK_WAITING;
	hook_init_count++;
	return;
  } else 
  // 2. run_hook_after()
  if (sim_parser->match("after")) {
	int t;
	sim_parser->scan_int(&t);
  	hook_after_table[hook_after_count].time = t;
	sim_parser->rest(hook_after_table[hook_after_count].command);
	hook_after_table[hook_after_count].exec_flag = SIM_HOOK_WAITING;
	hook_after_count++;
	return;
  } else 
  // 3. run_hook_before() 
  if (sim_parser->match("before")) {
	int t;
	sim_parser->scan_int(&t);
  	hook_before_table[hook_before_count].time = t;
	sim_parser->rest(hook_before_table[hook_before_count].command);
	hook_before_table[hook_before_count].exec_flag = SIM_HOOK_WAITING;
	hook_before_count++;
	return;
  }
  
}

//----------------------------------------------------------------------------
//	run_hooks_init() 
//	  - run commands that need to be run at the very beginning
//----------------------------------------------------------------------------
void simenv::run_hooks_init() {

  cout << "run_hooks_init" << endl;

  for (int i=0; i<hook_init_count; ++i) {

	if (hook_init_table[i].exec_flag == SIM_HOOK_WAITING) {
		parse(hook_init_table[i].command);
		cout << "simenv::run_hooks_init() : "
		     << "executing " << hook_init_table[i].command << endl;
		hook_init_table[i].exec_flag = SIM_HOOK_EXECED;
	}

  }

}

//----------------------------------------------------------------------------
//	run_hooks_after() 
//	  - check and run commands for time t
//----------------------------------------------------------------------------
void simenv::run_hooks_after(int t) {

  // check if it is time and run the hooks

  for (int i=0; i<hook_after_count; ++i) {

	if (hook_after_table[i].exec_flag == SIM_HOOK_WAITING
	    	&& hook_after_table[i].time == t) {
		parse(hook_after_table[i].command);
		cout << "simenv::run_hooks_after() : "
		     << "executing " << hook_after_table[i].command << endl;
		hook_after_table[i].exec_flag = SIM_HOOK_EXECED;
	}
  }

}

//----------------------------------------------------------------------------
//	run_hooks_before() 
//	  - check and run commands for time t
//----------------------------------------------------------------------------
void simenv::run_hooks_before(int t) {

  // check if it is time and run the hooks

  for (int i=0; i<hook_before_count; ++i) {

	if (hook_before_table[i].exec_flag == SIM_HOOK_WAITING
	        && hook_before_table[i].time == t) {
		cout << "simenv::run_hooks_before() : "
		     << "executing " << hook_before_table[i].command << endl;
		parse(hook_before_table[i].command);
		hook_before_table[i].exec_flag = SIM_HOOK_EXECED;
	}
  }

}

//----------------------------------------------------------------------------
//	list_hook() 
//	  - list all hooks
//----------------------------------------------------------------------------
void simenv::list_hook(SIM_HOOK_STAT stat) {


  // init hook
  cout << "[hook init list]" << endl;
  for (int i=0; i<hook_init_count; ++i) {
	if (hook_init_table[i].exec_flag == stat || stat == SIM_HOOK_ALL) {
		cout << "hook init "
		     << hook_init_table[i].command << endl;
	}
  }
  // after hook
  cout << "[hook after list]" << endl;
  for (int i=0; i<hook_after_count; ++i) {
	if (hook_after_table[i].exec_flag == stat || stat == SIM_HOOK_ALL) {
		cout << "hook after "
		     << hook_after_table[i].time << " "
		     << hook_after_table[i].command << endl;
	}
  }
  // before hook
  cout << "[hook before list]" << endl;
  for (int i=0; i<hook_before_count; ++i) {
	if (hook_before_table[i].exec_flag == stat || stat == SIM_HOOK_ALL) {
		cout << "hook before "
		     << hook_before_table[i].time << " "
		     << hook_before_table[i].command << endl;
	}
  }
}		     

//----------------------------------------------------------------------------
//	help() 
//	  - print all help_strings of commands
//----------------------------------------------------------------------------
void simenv::help() {

  char *line = "-----------------------------------"
	       "-----------------------------------------";
  cerr << endl;

  // list variables
  cerr << "simenv::help()" << endl;
  cerr << line << endl;
  cerr << "Global scalar variables" << endl;
  cerr << line << endl;
  list_scalar();

  for (int i=0; i<obj_count; ++i) { 
  	cerr << line << endl;
  	cerr << "Local scalar variables: " << obj_table[i].name << endl;
  	cerr << line << endl;
	if (obj_table[i].objp->net_bb!=NULL) {
        	obj_table[i].objp->net_bb->list_scalar();
	}
  }

  // list commands
  cerr << line << endl;
  cerr << "Registered commands : command [<Type:variable_name> ... ] " << endl;
  cerr << "	<Type> = [R|I] where R: real, I: integer. ] " << endl;
  cerr << " 	Number of commands = " << cmd_count << endl;
  cerr << line << endl;	

  for (int i=0; i<cmd_count; ++i) {
	cerr 	<< "[" << i << "] " << "run " << cmd_table[i].name << " " 
		<<  cmd_table[i].help_str << endl;
  }

  cerr << endl;

}

//----------------------------------------------------------------------------
//	save() 
//	  - save -- blackboard only
//----------------------------------------------------------------------------
int simenv::save(char *filename) {

  cerr << "simenv::save() : this funciton does nothing." << endl;

  return 1;

}


//----------------------------------------------------------------------------
//	load() 
//	  - blackboard only
//----------------------------------------------------------------------------
int simenv::load(char *filename) {

  cerr << "simenv::load() : this funciton does nothing." << endl;

  return 1;

}

//----------------------------------------------------------------------------
//	find_param() 
//	  - find parameter and return index. Return -1 if not found.
//----------------------------------------------------------------------------
int simenv::find_param(char *name) {

  // reuse blackboard routine.
  return find(name);

}


//----------------------------------------------------------------------------
//	find_obj() 
//	  - find object and return index. Return -1 if not found.
//----------------------------------------------------------------------------
int simenv::find_obj(char *name) {

  // brute-force search : I hate it!

  for (int i=0; i<obj_count; ++i) {
        if (strcmp(name,obj_table[i].name) == 0) {
                return (i);
        }
  }
  warning("simenv::find_obj","object \"%s\" not found",name);
  return -1;

}

//----------------------------------------------------------------------------
//	find_cmd() 
//	  - find command and return index. Return -1 if not found.
//----------------------------------------------------------------------------
int simenv::find_cmd(char *name) {

  // brute-force search : I hate it!

  for (int i=0; i<cmd_count; ++i) {
        if (strcmp(name,cmd_table[i].name) == 0) {
                return (i);
        }
  }
  return -1;

}

//----------------------------------------------------------------------------
//	list() 
//	  - list all parameters, objects, and commands
//----------------------------------------------------------------------------
void simenv::list() {

  cerr << "simenv::list() : unimplemented" << endl;

}

//----------------------------------------------------------------------------
//	init_obj() 
//	  - init a single object table entry
//----------------------------------------------------------------------------
void simenv::init_obj(char *name, SIM_OBJ *obj){

  // sanity check
  if (obj_count>=max) {
        error_exit("simenv:init_obj",
		"cannot add object to object blackboard : count reached max");
  }

  if (strlen(name)>BB_NAME_LEN)
        error_exit("simenv::init_obj","object name too long");

  strncpy(obj->name,name,strlen(name)+1);

}

//----------------------------------------------------------------------------
//	init_cmd() 
//	  - init a single command table entry
//----------------------------------------------------------------------------
void simenv::init_cmd(char *name, SIM_CMD *cmd, int num_args,
	char *help_str){

  // sanity check
  if (cmd_count>=max) {
        error_exit("simenv:init_cmd",
		"cannot add cmd to command blackboard"
		" : count reached max");
  }

  if (strlen(name)>BB_NAME_LEN)
        error_exit("simenv::init_cmd","command name too long");

  if (strlen(help_str)>SIM_STR_LEN)
        error_exit("simenv::init_cmd","help string too long");

  strncpy(cmd->name,name,strlen(name)+1);
  cmd->num_args	= num_args;
  strncpy(cmd->help_str,help_str,strlen(help_str)+1);

}

//----------------------------------------------------------------------------
//	rerun_hooks() 
//	  - go through 0--t and redo all hooks
//----------------------------------------------------------------------------
void simenv::rerun_hooks(int t) {

  // rerun all previously defined hooks : this should be run before
  //  loading the network image.
  for (int i=0; i<=t; ++i) {

	run_hooks_before(i);
	run_hooks_after(i);

  }

}
